@@ -167,7 +167,8 @@ class llvmIRrep():
         else:
             self._edits = list(edits)
             self._serialized_edits = decode_edits(edits)
-            self.sort_serialized_edits()
+            # TODO: remove sorting after PACT submission (both here and llvm-mutate.py)
+            # self.sort_serialized_edits()
             self.update_src_from_edits()
 
     @property
